# intrnglist.tcl --
#
#	This file contains code for manipulating sets of integers
#	represented as Tcl lists of integer ranges.
#
# Copyright (c) 1999, Neil McKay
#
# See the file "license.terms" for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# {$Header}

package provide intrnglist 1.0

namespace eval intrnglist {
    namespace export normalize combine contains complement
    namespace export extent remove numElements
}

# A properly-formatted integer range list is a Tcl list whose elements
# are either single integers or integer pairs, i.e. lists containing
# two integers. An integer range list is "normalized", or in simplest
# form, if it has the following properties:
#
#	1) The ranges in the list are non-degenerate, i.e. the
#	   minimum value is less than or equal to the maximum value.
#	   (Reason: a degenerate range can be deleted without changing
#	   the set.)
#
#	2) All ranges that contain a single integer (those where the
#	   maximum and minimum values are equal) are represented as
#	   a single integer. (Reason: this shortens the list without
#	   changing the set.)
#
#	3) None of the ranges in the list overlap. (Reason: overlapping
#	   ranges can be combined, giving a shorter representation
#	   of the same set.)
#
#	4) None of the ranges in the list are adjacent, i.e. the
#	   minimum value in one range is not equal to one plus the
#	   maximum value of another range. (Reason: adjacent ranges
#	   can be combined without changing the meaning of the set.)
#
#	5) The ranges are sorted in ascending order. (Reason: this
#	   simplifies certain computations, and does not decrease
#	   the generality of the data representation.)
#

# intrnglist:normalize --
#
#	Makes sure that a list of integer ranges is normalized.
#
# Arguments:
#	l	Integer range list
#
# Results:
#	Returns a normalized version of the input list.

proc intrnglist::normalize {l} {
    # Don't do anything if we don't have to.

    if {[llength $l] == 0} {
	    return $l
    }

    # First, sort the intervals based on the lower
    # bound of the interval.

    set l [lsort -integer -index 0 $l]

    # Find the first non-empty interval.

    set nIntervals [llength $l]
    for {set i 0} {$i < $nIntervals} {incr i} {
	set curInterval [lindex $l $i]
	set curLow [lindex $curInterval 0]
	set curHigh [lindex $curInterval end]
	if {$curLow <= $curHigh} {
	    break
	}
    }
    incr i
    set l [lrange $l $i end]

    # For all the remaining intervals, merge in the next interval
    # if possible; if not, then write out the interval and try
    # to merge in the following one.

    set normList {}
    foreach interval $l {
	set low [lindex $interval 0]
	set high [lindex $interval end]
	if {$low > $high} {
	    continue
	}

	if {$low <= $curHigh + 1} {
	    if {$high > $curHigh} {
		set curHigh $high
	    }
	} else {
	    if {$curLow == $curHigh} {
		lappend normList [list $curLow]
	    } else {
		lappend normList [list $curLow $curHigh]
	    }
	    set curLow $low
	    set curHigh $high
	}
    }

    # Append the last interval.

    if {$curLow == $curHigh} {
	lappend normList [list $curLow]
    } else {
	lappend normList [list $curLow $curHigh]
    }

    # Done!

    return $normList
}

# intrnglist::combine --
#
#	Combine two or more integer range lists, i.e. perform
#	a set union on them.
#
# Arguments:
#	l		Integer range list
#
#	(others)	more integer range lists to union with "l"
#
# Results:
#	Returns a normalized integer range list that is the union
#	of all the arguments.

proc intrnglist::combine {l args} {
    foreach ll $args {
	set l [concat $l $ll]
    }
    return [normalizeIntRangeList $l]
}

# intrnglist::contains --
#
#	Determine whether an integer range list contains a given integer
#
# Arguments:
#	l	Integer range list
#
#	num	number to test for containment
#
# Results:
#	Returns a Boolean: 1 if "l" contains "num", 0 otherwise.


proc intrnglist::contains {l num} {
    foreach range $l {
	set low [lindex $range 0]
	set high [lindex $range end]
	if {$num >= $low && $num <= $high} {
	    return 1
	}
    }

    return 0
}

# intrnglist::complement --
#	Returns the complement of an integer range list
#	with respect to a given interval.
#
# Arguments:
#	l	List to take complement of (must be normalized)
#
#	min	Lower bound of interval to complement with respect to
#
#	max	Upper bound of interval to complement with respect to
#
# Results:
#	Returns the complemented set.

proc intrnglist::complement {l min max} {
    # Handle the trivial case first.

    if {[llength $l] <= 0} {
	    return [list [list $min $max]]
    }

    # "comp" is the complement set.

    set comp {}
    set prevHigh [expr {$min - 1}]

    # Scan down the list of intervals, producing intervals
    # representing the "gaps" between the intervals of "l".

    foreach range $l {
	set low [lindex $range 0]
	set high [lindex $range end]

	# Append the appropriate interval.

	incr low -1
	if {$low > $max} {
	    set low $max
	}
	incr prevHigh
	if {$prevHigh < $min} {
	    set prevHigh $min
	}
	lappend comp [list $prevHigh $low]

	# Next...

	set prevHigh $high
    }

    # Take care of the final interval.

    incr prevHigh
    lappend comp [list $prevHigh $max]

    # Normalize everything, and we're done.

    return [normalizeIntRangeList $comp]
}

# intrnglist::extent --
#
#	Generates the "extent" of an integer range list, i.e.
#	the smallest single interval containing the set.
#
# Arguments:
#	l	List to find the extent of
#
# Results:
#	 Returns the extent of a range list, i.e. a list containing
#	the minimum and maximum values in the set. If "l" is the
#	empty list, then the empty list is returned.

proc intrnglist::extent {l} {
    if {[llength $l] == 0} {
	return {}
    }

    set range [lindex $l 0]
    set l [lrange $l 1 end]
    set min [lindex $range 0]
    set max [lindex $range end]

    foreach range $l {
	set low [lindex $range 0]
	set high [lindex $range end]

	if {$low < $min} {
	    set min $low
	}

	if {$high > $max} {
	    set max $high
	}
    }

    return [list $min $max]
}

# intrnglist::remove --
#
#	Removes the integers contained in one set from another, i.e.
#	this performs set subtraction on integer range lists.
#
# Arguments:
#	l		Integer range list that we'll be removing
#			elements from
#
#	(others)	Sets to be subtracted from "l"
#
# Results:
#	Returns the result of the set subtraction.

proc intrnglist::remove {l args} {
    # Deal with the easy cases first.

    if {[llength $l] == 0} {
	return {}
    }

    if {[llength $args] == 0} {
	return $l
    }

    # We'll compute the result using DeMorgan's laws; we use the
    # identity L*(~B) == ~((~L)+B), where "~" is set complement,
    # "*" is intersection, and "+" is union. First, we have to
    # compute the single set we're subtracting; it's just the union
    # of all the sets we're subtracting.

    set b {}
    foreach s $args {
	set b [concat $b $s]
    }
    set b [normalizeIntRangeList $b]

    # Now we'll compute the range with which we'll compute complements.

    set compRange [extentIntRangeList $l]
    set min [lindex $compRange 0]
    set max [lindex $compRange 1]

    # Now do it...

    set lComp [complementIntRangeList $l $min $max]
    return [complementIntRangeList [combineIntRangeLists $lComp $b] $min $max]
}

# intrnglist::numElements --
#
#	Finds the number of elements in a set of integers
#	represented as a range list
#
# Arguments:
#	l		Integer range list whose elements
#			we'll be counting
#
# Results:
#	Returns the number of elements in the set

proc intrnglist::numElements {l} {
    set n 0
    foreach range $l {
	set low [lindex $range 0]
	set high [lindex $range end]
	set n [expr {$n + $high - $low + 1}]
    }

    return $n
}
