# newsrc.tcl --
#
#	This file contains code for manipulating ".newsrc" files.
#
# Copyright (c) 1999, Neil McKay
#
# See the file "license.terms" for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# {$Header}

package provide newsrc 1.0

namespace eval newsrc {
    namespace export readFile writeFile
}

# newsrc::readFile --
#	Reads a file in the ".newsrc" format
#
# Arguments:
#	fileHandle	File handle to read from
#
# Results:
#	Returns a Tcl list. The list consist of a newsgroup name
#	followed by information about that newsgroup, followed by
#	another newsgroup name, followed by more information, etc.
#	The information about the newsgroup consists of a two-element
#	list: a Boolean indicating whether the user is currently
#	subscribed to this newsgroup, followed by a list of article-number
#	ranges. The article number ranges may be either a single
#	number, or a pair of numbers that indicate the range.

proc newsrc::readFile {fileHandle} {
    # Read lines until we hit end-of-file.

    while {![eof $fileHandle]} {
	set nChars [gets $fileHandle line]
	if {$nChars < 0} {
	    break
	}

	# Skip blank lines.

	set line [string trim $line]
	if {[string compare $line ""] == 0} {
	    continue
	}

	# Figure out where the newsgroup name ends.

	set idx [string first ":" $line]
	if {$idx < 0} {
	    set idx [string first "!" $line]
	}

	if {$idx < 0} {
	    return -code error \
			{Badly formatted line in newsrc file: no ":" or "!"}
	}

	# Extract out the newsgroup name.

	set newsgroup [string range $line 0 [expr {$idx - 1}]]

	# Are we subscribed to this, or not?

	switch [string index $line $idx] {
	    ":" {
		set subscribed 1
	    }
	    "!" {
		set subscribed 0
	    }
	}

	# Get the list of article number ranges.

	set line [string trim [string range $line [expr {$idx + 1}] end]]
	set rawRangeList [split $line ","]
	set rangeList {}
	foreach range $rawRangeList {
	    lappend rangeList [split $range "-"]
	}

	# Put info into the info array.

	set infoArray($newsgroup) [list $subscribed $rangeList]
    }

    return [array get infoArray]
}

# newsrc::writeFile --
#
#	Writes information about subscribed newsgroups and articles
#	read, in the format of a ".newsrc" file.
#
# Arguments:
#	fileHandle	File handle to write newsgroup info to
#
#	info		Newsgroup information, in the format returned
#			by the "readFile" proc
#
# Results:
#	None, or an error message

proc newsrc::writeFile {fileHandle info} {
    array set infoArray $info
    foreach newsgroup [lsort [array names infoArray]] {
	    set info $infoArray($newsgroup)

	    set subscribed [lindex $info 0]
	    set subscribedSym "!"
	    if {$subscribed} {
		set subscribedSym ":"
	    }

	    set rangeList [lindex $info 1]
	    set processedRangeList {}
	    foreach range $rangeList {
		lappend processedRangeList [join $range "-"]
	    }
	    set processedRangeList [join $processedRangeList ","]

	    puts $fileHandle "${newsgroup}${subscribedSym} $processedRangeList"
    }
}

