# nntp.tcl --
#
#	This file contains code for doing NNTP (network news) transactions
#
# Copyright (c) 1999, Neil McKay
#
# See the file "license.terms" for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# {$Header}

package provide nntp 1.0

namespace eval nntp {
    namespace export openConnection setResponseHandler getResponseHandler
    namespace export responseCode closeConnection postingAllowed
    namespace export getGroupList getNewGroups setGroup
    namespace export getHeader getBody getNewNews setArticlePointer
    namespace export nextArticle prevArticle post clientIsSlave iHave

    # These are arrays for keeping track of information related
    # to NNTP connections. They are indexed by a "connection
    # handle", which is a unique string associated with each
    # NNTP connection. "nntpSocket" contains the name of the
    # socket used for communication with the server.
    # "nntpResponseHandler" contains the name of a procedure that
    # is passed the server's reply messages. "serverAllowsPosting"
    # contains a Boolean indicating whether the server allows
    # clisents to post messages to it.

    variable nntpSocket
    variable nntpResponseHandler
    variable serverAllowsPosting
}

# nntp::responseCode --
#
#	Extract the numeric response code from the response
#	message returned by the NNTP server.
#
# Arguments:
#	msg	Response message returned by NNTP server
#
# Results:
#	Returns the numeric response code (the first "word"
#	of the server's response message)

proc nntp::responseCode {msg} {
    set msg [string trimleft $msg]
    return [lindex [split $msg " "] 0]
}


# nntp::openConnection --
#
#	Try to open a connection to an NNTP server.
#
# Arguments:
#
#	host	Name of host where NNTP server is running
#
# Options:
#
#	-port <portNo>		TCP port to connect to; defaults to 119
#
#	-responsehandler <hnd>	Server response handler routine; defaults
#				to the empty string, which means do nothing
#				with the NNTP server response messages. <hnd>
#				should be a proc that accepts a single
#				argument: the server response message.
#
#	-user <userName>	Use <userName> as the user name for
#				authentication
#
#	-pass <passwd>		Use <passwd> as the password for
#				authentication
#
# Results:
#
#	If the open succeeds, it returns an "nntp connection handle",
#	which is a unique ID associated with the connection. If it fails,
#	it returns an error message, and returns an error code of 1.

proc nntp::openConnection {host args} {
    variable nntpSocket
    variable nntpResponseHandler
    variable serverAllowsPosting

    # Set defaults for optional arguments, and parse the options.

    set opts(-port) 119
    set opts(-responsehandler) {}

    array set opts $args

    # Generate a unique handle name.

    set handleNum 0
    while {1} {
	set nntpHandle "nntphandle$handleNum"
	if {![info exists nntpSocket($nntpHandle)]} {
	    break
	}
	incr handleNum
    }

    # Try to open the NNTP connection.

    set ecode [catch [list socket $host $opts(-port)] sock]

    if {$ecode != 0} {
	return -code error "Couldn't make news connection: $sock"
    }

    # Got it! Save the socket, and set the message handling routine.

    set nntpSocket($nntpHandle) $sock
    set responseHandler $opts(-responsehandler)
    set nntpResponseHandler($nntpHandle) $responseHandler

    # Configure the socket so that it trims off the trailing CRLF.

    fconfigure $sock -translation crlf

    # Get the server's response string.

    set nChars [gets $sock response]

    if {$nChars < 0} {
	catch [list ::close $sock]
	unset nntpSocket($nntpHandle)
	unset nntpResponseHandler($nntpHandle)
	return -code error "End-of-file on socket"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Note whether the server allows posting.

    set resp [responseCode $response]
    set serverAllowsPosting($nntpHandle) [expr {$resp == 200}]

    # Handle the authorization stuff, if necessary.

    if [info exists opts(-user)] {
	puts $sock "AUTHINFO USER $opts(-user)"
	flush $sock

	set nchars [gets $sock response]
	if {$nchars < 0} {
	    catch [list ::nntp::close $nntpHandle]
	    return -code error "AUTHINFO USER fails"
	}
    }

    if [info exists opts(-pass)] {
	puts $sock "AUTHINFO PASS $opts(-pass)"
	flush $sock

	set nchars [gets $sock response]
	if {$nchars < 0} {
	    catch [list ::nntp::close $nntpHandle]
	    return -code error "AUTHINFO PASS fails"
	}
    }

    # Return the connection handle.

    return $nntpHandle
}

# nntp::postingAllowed --
#
#	Indicate whether the NNTP server allows posting of articles
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	1 if posting is allowed, 0 if not

proc nntp::postingAllowed {nntpHandle} {
    variable serverAllowsPosting

    return $serverAllowsPosting($nntpHandle)
}

# nntp::setResponseHandler --
#
#	Set the message handling function associated with the
#	given server connection
#
# Arguments:
#	nntpHandle		NNTP server connection handle
#
#	responseHandlerProc	Proc that will be executed when
#				the server sends back a response
#				message. This proc must accept
#				one argument: the response message.
#
# Results:
#	None

proc nntp::setResponseHandler {nntpHandle responseHandlerProc} {
    variable nntpResponseHandler

    set nntpResponseHandler($nntpHandle) $responseHandlerProc
}


# nntp::getResponseHandler --
#
#	Get the message handling function associated with the
#	given server connection
#
# Arguments:
#	nntpHandle		NNTP connection handle
#
# Results:
#	returns the name of the response handler proc

proc nntp::getResponseHandler {nntpHandle} {
    variable nntpResponseHandler

    return $nntpResponseHandler($nntpHandle)
}

# closeConnection --
#
#	Close the connection to an NNTP server
#
# Arguments:
#	nntpHandle	NNTP connection handle
#
# Results:
#	None

proc nntp::closeConnection {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler
    variable serverAllowsPosting

    # Try to send a QUIT command; then close the socket, and clean up.

    set sock $nntpSocket($nntpHandle)
    puts $sock "QUIT"
    ::flush $sock
    ::close $sock
    unset nntpSocket($nntpHandle)
    catch [list unset nntpResponseHandler($nntpHandle)]
    catch [list unset serverAllowsPosting($nntpHandle)]
}

# nntp::ReadGroupList --
#
#	Reads a newsgroup list from the server. It assumes
#	that all the preliminary setup has been done, and that the
#	client only needs to read the newsgroup list and return it.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	Returns a list of newsgroup info, in the form of a
#	Tcl list. The elements of the list come in groups of
#	four: the newsgroup name, the article number of the first
#	article, the article number of the last article, and the
#	posting status for the newsgroup.

proc nntp::ReadGroupList {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    set newsgroupList {}
    while {1} {
	# Get a line of text; we're done if it consists of a single ".".

	set nchars [gets $sock line]
	if {$nchars < 0} {
	    catch [list closeConnection $nntpHandle]
		return -code error "Server disconnected prematurely"
	}

	if {[string compare $line "."] == 0} {
	    return $newsgroupList
	}

	# If the first char is a ".", then trim it off, since
	# an initial "." will be doubled.

	set firstChar [string index $line 0]
	if {[string compare $firstChar "."] == 0} {
	    set line [string range $line 1 end]
	}

	# Extract the newsgroup name, first article number,
	# last article number, and the post status.
	# NOTE: we really shouldn't use "lindex" to do this,
	# but this quick-and-dirty technique does seem to work...

	set groupName [lindex $line 0]
	scan [lindex $line 1] "%d" firstArticle
	scan [lindex $line 2] "%d" lastArticle
	set postStatus [lindex $line 3]
	lappend newsgroupList $groupName \
			$firstArticle $lastArticle $postStatus
    }
}

# nntp::getGroupList --
#
#	Retrieve a list of newsgroup names from an NNTP server
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	Returns a list of newsgroups, in the same format as
#	ReadGroupList.

proc nntp::getGroupList {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    puts $sock "LIST"
    flush $sock

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    if {[responseCode $response] != 215} {
	return -code error "Newsgroups not retrieved"
    }

    return [ReadGroupList $nntpHandle]
}

# nntp::getNewGroups --
#
#	Retrieves a list of newsgroup names from the server associated
#	with "nntpHandle"; newsgroup names are only reported if the
#	newsgroup has news that is newer than a specified date
#	and time.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	datime		Date and time, specified as an integer
#			in the format of the Tcl "clock" command.
#			Newsgroup names will be listed only if
#			they have received articles after this
#			date and time.
#
# Options:
#	-distlist <dists>	Restricts what "distributions"
#				will be reported, i.e. what
#				top-level newsgroup names (like
#				:net", "alt", "comp", etc.) will
#				be looked at.
#
#	-gmt <Boolean>		If "Boolean" is non-zero, then
#				"datime" will be interpreted as
#				UTC (Univeral time, or Greenwich
#				mean time); otherwise, it uses
#				local time. (Local time is the default.)
#
# Results:
#	Returns a list of newsgroup names, in the format
#	of ReadGroupList.

proc nntp::getNewGroups {nntpHandle datime args} {
    variable nntpSocket
    variable nntpResponseHandler

    set opts(-gmt) 0
    set opts(-distlist) {}

    array set opts $args

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Format the date/time properly, and put it
    # into the NEWGROUPS command line.

    set datimeFmt  [clock format $datime \
			-format "%y%m%d %H%M%S" -gmt $opts(-gmt)]
    set cmd "NEWGROUPS $datimeFmt"
    if {$opts(-gmt)} {
	append cmd " GMT"
    }

    # Format the distribution list properly.

    if {[llength $opts(-distlist)] != 0} {
	set fmtDistrib [join $opts(-distlist) ","]
	append cmd " <$fmtDistrib>"
    }

    # Send the command to the server.

    puts $sock $cmd
    flush $sock

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    if {[responseCode $response] != 231} {
	return -code error "Newsgroups not retrieved"
    }

    return [ReadGroupList $nntpHandle]
}

# nntp::setGroup --
#
#	Sets the current newsgroup for a specified server connection
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	groupName	Name of newsgroup to set as "current"
#			for the server connection
#
# Results:
#	If successful, returns a three-element list:
#	the newsgroup name, and the numbers of the  first
#	and last articles. If unsuccessful, returns an
#	error message, and a return code of 1.

proc nntp::setGroup {nntpHandle groupName} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "GROUP" command, and get a response.

    puts $sock "GROUP $groupName"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 211} {
	return -code error "Error setting newsgroup: $response"
    }

    # We're OK: return the info.

    set groupName [lindex $response 4]
    set firstArticleNo [lindex $response 2]
    set lastArticleNo [lindex $response 3]
    return [list $groupName $firstArticleNo $lastArticleNo]
}

# nntp::ReadMessageText --
#
#	Reads a message from the specified server connection
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	If successful, returns text read from the server. Otherwise,
#	it returns an error message, with an error code of 1.

proc nntp::ReadMessageText {nntpHandle} {
    variable nntpSocket

    set sock $nntpSocket($nntpHandle)
    set text ""
    while {1} {
	# Get a line of text; we're done if it consists of a single ".".

	set nchars [gets $sock line]
	if {$nchars < 0} {
	    catch [list closeConnection $nntpHandle]
	    return -code error "Server disconnected prematurely"
	}

	if {[string compare $line "."] == 0} {
	    return $text
	}

	# If the first char is a ".", then trim it off, since
	# an initial "." will be doubled.

	set firstChar [string index $line 0]
	if {[string compare $firstChar "."] == 0} {
	    set line [string range $line 1 end]
	}

	# Add the line...

	append text $line "\n"
    }
}

# nntp::getHeader --
#
#	Reads the header of a server connection's currently-
#	selected article.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	If successful, returns the header text. If not successful,
#	returns an error message, with a return code of 1.

proc nntp::getHeader {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "HEAD" command, and get a response.

    puts $sock "HEAD"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 221} {
	return -code error "Error reading article header: $response"
    }

    # Return the header text.

    return [ReadMessageText $nntpHandle]
}

# nntp::getBody --
#
#	Reads the body of a server connection's currently-
#	selected article.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	If successful, returns the body text. If not successful,
#	returns an error message, with a return code of 1.

proc nntp::getBody {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "BODY" command, and get a response.

    puts $sock "BODY"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 222} {
	return -code error "Error reading article text: $response"
    }

    # Return the message body.

    return [ReadMessageText $nntpHandle]
}


# nntp::getNewNews --
#
#	Retrieves "new" news from an nntp server.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	groupList	List of newsgroups we're interested
#			in. This is a Tcl list; the elements
#			of the list may contain "*" wildcard
#			characters, and may be preceded by
#			"!" to indicate groups we're NOT
#			interested in.
#
#	datime		Date and time, given as an integer,
#			in the format of the "clock" command.
#
# Options:
#	-gmt <Bool>		If <Bool> is non-zero, use
#				Greenwich Mean Time; otherwise,
#				use local time. (Default is
#				local time.)
#
#	-grouplist <list>	specifies list of newsgroup
#				specifiers; like the "-distlist"
#				argument of the getNewGroups
#				command
#
# Results:
#	Returns the result of the NEWNEWS command as a Tcl list,
#	with each line of the result as a list element.

proc nntp::getNewNews {nntpHandle groupList datime args} {
    variable nntpSocket
    variable nntpResponseHandler

    set opts(-gmt) 0
    set opts(-grouplist) {}

    array set opts $args

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Format the date/time properly, and put it
    # into the NEWNEWS command line.

    set datimeFmt  [clock format $datime \
				-format "%y%m%d %H%M%S" -gmt $opts(-gmt)]
    set cmd "NEWNEWS [join $groupList ","] $datimeFmt"
    if {$opts(-gmt)} {
	    append cmd " GMT"
    }

    # Format the group list properly.

    if {[llength $opts(-grouplist)] != 0} {
	    set fmtGroups [join $opts(-grouplist) ","]
	    append cmd " <$fmtGroups>"
    }

    # Send the command to the server.

    puts $sock $cmd
    flush $sock

    set nchars [gets $sock response]
    if {$nchars < 0} {
	    catch [list closeConnection $nntpHandle]
	    return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	    $responseHandler $response
    }

    if {[responseCode $response] != 230} {
	    return -code error "Newsgroups not retrieved"
    }

    return [split [ReadMessageText $nntpHandle] "\n"]
}

# nntp::setArticlePointer --
#	Sets the current article pointer for a given NNTP server
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	article		Number or ID of article to point to
#
# Results:
#	None, unless an error occurs.

proc nntp::setArticlePointer {nntpHandle article} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "STAT" command, and get a response.

    puts $sock "STAT $article"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 223} {
	return -code error "Error setting article pointer: $response"
    }

    return
}

# nntp::nextArticle --
#
#	Advances the article pointer, i.e. moves the pointer
#	to the next article.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	Returns the (new) current article number, or an error.

proc nntp::nextArticle {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "NEXT" command, and get a response.

    puts $sock "NEXT"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 223} {
	return -code error "Error going to next article: $response"
    }

    # Return the article number.

    return [lindex $response 1]
}

# nntp::prevArticle --
#
#	Decrements the article pointer, i.e. moves the pointer
#	to the previous article.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	Returns the (new) current article number, or an error.

proc nntp::prevArticle {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "LAST" command, and get a response.

    puts $sock "LAST"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 223} {
	return -code error "Error going to next article: $response"
    }

    # Return the article number.

    return [lindex $response 1]
}

# nntp::post --
#
#	Posts an article on a given server
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	header		Header of article
#
#	body		Body of article
#
# Results:
#	None, or an error.

proc nntp::post {nntpHandle header body} {
    variable nntpSocket
    variable nntpResponseHandler
    variable serverAllowsPosting

    # Make sure we're allowed to post.

    if {!$serverAllowsPosting($nntpHandle)} {
	return -code error "Server does not allow posting"
    }

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "POST" command.

    puts $sock "POST"
    flush $sock

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    if {[responseCode $response] != 340} {
	return -code error "Could not post message"
    }

    # Write the message: header first, then body.

    set text "$header\n\n$body"
    foreach line [split $text "\n"] {
	# If the line starts with a ".", send an extra "."

	set firstChar [string index $line 0]
	if {[string compare $firstChar "."] == 0} {
	    set line ".$line"
	}

	puts $sock $line
    }

    # End with a single "." on a line.

    puts $sock "."
    flush $sock

    # Get the response.

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    if {[responseCode $response] != 240} {
	return -code error "Posting failed"
    }

    return
}

# nntp::clientIsSlave --
#	Indicates to the server that the client is a "slave",
#	i.e. it is a slave server requesting a newsfeed, rather
#	than a normal interactive user.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
# Results:
#	None, or an error message.

proc nntp::clientIsSlave {nntpHandle} {
    variable nntpSocket
    variable nntpResponseHandler

    set sock $nntpSocket($nntpHandle)
    set responseHandler $nntpResponseHandler($nntpHandle)

    # Send the "SLAVE" command, and get a response.

    puts $sock "SLAVE"
    flush $sock
    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    # Check for errors...

    set msgCode [responseCode $response]
    if {$msgCode != 202} {
	return -code error "Error going to next article: $response"
    }

    return {}
}

# nntp::iHave --
#	Indicates to the server that the client has a particular
#	message, and sends the message if the server want it.
#
# Arguments:
#	nntpHandle	NNTP server connection handle
#
#	messageID	ID of message to be sent to server
#
#	header		Header of message to be sent to server
#
#	body		Body of message to be sent to server
#
# Results:
#	None, or an error message.

proc nntp::iHave {nntpHandle messageID header body} {
    variable nntpSocket
    variable nntpResponseHandler
    variable serverAllowsPosting

    # Make sure we're allowed to post.

    if {!$serverAllowsPosting($nntpHandle)} {
	return -code error "Server does not allow posting"
    }

    set sock $nntpSocket($nntpHandle)

    # Send the "IHAVE" command.

    puts $sock "IHAVE <$messageID>"
    flush $sock

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    set rcode [responseCode $response]
    if {$rcode == 435} {
	return
    }

    if {$rcode != 335} {
	return -code error "Could not transfer article"
    }

    # Write the message: header first, then body.

    set text "$header\n\n$body"
    foreach line [split $text "\n"] {
	# If the line starts with a ".", send an extra "."

	set firstChar [string index $line 0]
	if {[string compare $firstChar "."] == 0} {
	    set line ".$line"
	}

	puts $sock $line
    }

    # End with a single "." on a line.

    puts $sock "."
    flush $sock

    # Get the response.

    set nchars [gets $sock response]
    if {$nchars < 0} {
	catch [list closeConnection $nntpHandle]
	return -code error "Server disconnected prematurely"
    }

    if {[string compare $responseHandler ""] != 0} {
	$responseHandler $response
    }

    if {[responseCode $response] != 235} {
	return -code error "Article transfer failed"
    }
}

